#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>

#define MAX_BUF    50
#define SLEEP_TIME 1
#define DEF_SIGNAL "TERM"

static void child1(void);
static void child2(pid_t child1, char *signal);
static void child_handler(int signum);
static int wait_all(int n);

int main(int argc, char **argv) {
  pid_t pid;
  char *sig;
  sig = (argc == 2) ? argv[1] : DEF_SIGNAL;
  /* Create child 1 */
  pid = fork();
  switch(pid) {
  case 0:
	child1();
	return(EXIT_SUCCESS);
  case -1:
	perror("fork1");
	return(EXIT_FAILURE);
  }
  /* Create child 2 */
  switch(fork()) {
  case 0:
	child2(pid, sig);
	return(EXIT_FAILURE);
  case -1:
	perror("fork2");
	return(EXIT_FAILURE);
  }
  return (wait_all(2)) ? (EXIT_FAILURE) : (EXIT_SUCCESS);
}


static void child1(void) {
  /* Add signal handlers */
  signal(SIGUSR1, child_handler);
  signal(SIGUSR2, child_handler);
  signal(SIGTERM, child_handler);
  signal(SIGALRM, child_handler);
  alarm(SLEEP_TIME*3);
  /* Wait for signal */
  pause();
  exit(EXIT_SUCCESS);
}

static void child2(pid_t child1, char *signal) {
  char str_pid[MAX_BUF];
  sleep(SLEEP_TIME);
  snprintf(str_pid, MAX_BUF, "%d", (int)child1);
  printf("Child %d is calling 'kill -s %s %s'\n", (int)getpid(), signal, str_pid);
  execlp("kill", "kill", "-s", signal, str_pid, NULL);
  perror("execlp");
  exit(EXIT_FAILURE);
}

static int wait_all(int n) {
  pid_t pid;
  int status;
  while (n--) {
	if ((pid = wait(&status)) == -1) {
	  perror("wait");
	  exit(-1);
	}
	if (WIFEXITED(status)) {
	  printf("Child %d terminated with exit status %d.\n", 
			 (int)pid, WEXITSTATUS(status));
	}
	else {
	  /* Returns true if the child process was terminated by a signal */ 
	  if (WIFSIGNALED(status)) {
		printf("Child %d was terminated by the signal %d.\n",
			   (int)pid, WTERMSIG(status));
	  }
	  else {
		printf("Child %d terminated abnormally.\n", (int)pid);
	  }
	}
  }
  return(0);
}

static void child_handler(int signum) {
  char *sig_name;
  int value;
  switch(signum) {
  case SIGUSR1:
	sig_name = "USR1";
	value = 1;
	break;
  case SIGUSR2:
	sig_name = "USR2";
	value = 2;
	break;
  case SIGTERM:
	sig_name = "TERM";
	value = 3;
	break;
  case SIGALRM:
	printf("Alarm expired...\n");
	return;
	break;
  default:
	printf("Child %d receives signal %d: exit(5)\n", (int)getpid(), signum);
	exit(5);
  }
  printf("Child %d receives signal SIG%s: exit(%d)\n", 
		 (int)getpid(), sig_name, value);
  sleep(SLEEP_TIME);
  exit(value);
}
