#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>

#define MAX_BUF    50
#define MAX_SLEEP  15

static void parent_handler(int signum);
static int create_children(int n);

/* Created children */
static int children;

int main(int argc, char **argv) {
  int n, status;
  char buf[MAX_BUF];
  /* Add signal handler */
  signal(SIGTERM, parent_handler);
  signal(SIGINT, parent_handler);
  signal(SIGCHLD, parent_handler);
  /* Reset children number */
  children = 0;
  /* Loop */
  do {
	/* Read number of children */
	printf("-- Children: %d -- Insert the number of children:\n", children);
	if (!fgets(buf, MAX_BUF, stdin)) {
	  perror("Read line");
	  break;
	}
	n = atoi(buf);
	status = create_children(n);
	printf("Created %d children: %s\n", n, (status) ? "ERROR" : "OK");
  } while (!status);
  return(EXIT_FAILURE);
}

static int create_children(int n) {
  /* Create n children */
  while(n--) {
	switch(fork()) {
	case 0:
	  /* Child */
	  srand((unsigned int)getpid());
	  n = rand() % MAX_SLEEP;
	  printf("[Child %d] Sleep for %d seconds.\n", (int)getpid(), n);
	  sleep(n);
	  exit(n);
	case -1:
	  perror("fork");
	  return(-1);
	}
	/* Update children number */
	children++;
  }
  return 0;
}

static void parent_handler(int signum) {
  pid_t pid;
  int status;
  switch(signum) {
  case SIGTERM:
  case SIGINT:
	printf("Bye bye...\n");
	if (children) {
	  printf("Children: %d\n", children);
	  exit(EXIT_FAILURE);
	}
	exit(EXIT_SUCCESS);
  case SIGCHLD:
	if ((pid = wait(&status)) == -1) {
	  perror("wait");
	  return;
	}
	children--;
	if (WIFEXITED(status)) {
	  printf("Child %d terminated with status %d (Children %d).\n",
			 (int)pid, WEXITSTATUS(status), children);
	}
	else {
	  printf("Child %d terminated abnormally (Children %d).\n",
			 (int)pid, children);
	}
  }
}

